import { collect, Symbol, ws } from './common.js'
import { Node, Edge, KeyValuePair, Label, Name } from './types.js'
import { curry, join } from '../fn.js'
import { alpha, alphanumeric, many, maybe, map, parse, seq, skip, between, noCaseString, separated, until, list } from '../parser.js'

const { Bracket, Colon, Comma, Hyphen, Quote } = Symbol

const name = map(
  ([x]) => new Name(x),
  collect(seq(alpha, many(alphanumeric)))
)

const label = map(([x]) => new Label(x), seq(skip(Colon), name))

const trim = curry((parser, state) => (
  between(ws, parser, ws, state)
))

const bracketed = curry((value, { Left, Right }, state) => (
  between(trim(Left), value, trim(Right), state)
))

const quoted = map(join(''), seq(skip(Quote), until(Quote), skip(Quote)))
const kvp = map(
  ([k, v]) => new KeyValuePair(k, v),
  separated(name, trim(Colon), quoted)
)
const kvps = list(trim(Comma), kvp)
export const properties = bracketed(kvps, Bracket.Curly)

const id = seq(maybe(name), label)

export const node = map(
  ([name, label, ...properties]) => new Node(name, label, properties),
  bracketed(seq(id, ws, maybe(properties)), Bracket.Round)
)

export const edge = map(
  ([id]) => new Edge(id),
  bracketed(id, Bracket.Square)
)

const rightArrow = seq(Hyphen, Bracket.Angle.Right)
const relationship = seq(skip(Hyphen), edge, skip(rightArrow), node)

const keyword = noCaseString('match')
const params = seq(node, many(relationship))
export const statement = seq(skip(keyword), ws, params)

